package tutorial.dao;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.orm.jpa.support.JpaDaoSupport;
import org.springframework.transaction.annotation.Transactional;

import tutorial.model.ManagedObject;
import tutorial.model.PathProxy;
import tutorial.service.PathProxyService;
import tutorial.util.EntityPath;

/**
 * <b>Created:</b> Apr 19, 2008<br>
 * <b>Title:</b> PathProxyDao<br>
 * <b>Description:</b><br>
 * <b>Copyright:</b> Copyright (c) 2008<br>
 * @author MTyson
 * <p>
 * The implementation of PathProxyService.  This class holds all the core logic for dealing with PathProxy.
 */

public class PathProxyDao extends JpaDaoSupport implements PathProxyService {
    private static final transient Log log = LogFactory.getLog(tutorial.dao.PathProxyDao.class);
    
    /**default empty constructor*/
    public PathProxyDao() {
    }
    @Transactional
    public void addChild(EntityPath parentPath, ManagedObject child){
        PathProxy pp = new PathProxy();
        pp.setEntityId(child.getId());
        pp.setEntityType(child.getClass().getName());
        PathProxy parentPp = this.getPathProxyFromMoArray(parentPath.getPathAsObjects(), true);
        pp.setParent(parentPp);
        this.getJpaTemplate().persist(pp);
    }
    
    public List<Object> getPathChildren(EntityPath path, String type){
        PathProxy parentPp = this.getPathProxyFromMoArray(path.getPathAsObjects(), false);
        if (log.isTraceEnabled()) { log.trace("parentPp: " + parentPp); }
        List<Object> list = new ArrayList<Object>();
        if (parentPp == null){
            if (log.isInfoEnabled()) { log.info("No object found on path: " + ArrayUtils.toString(path)); }
        } else {
            String hql = "select t from " + type + " t, PathProxy pp where t.id = pp.entityId " +
                " and pp.parent = :parentPp";
            if (log.isTraceEnabled()) { log.trace("hql: " + hql); }
            Map<String, Object> params = new HashMap<String, Object>();
            params.put("parentPp", parentPp);
            list = this.getJpaTemplate().findByNamedParams(hql, params);
        }
        if (log.isInfoEnabled()) { log.info("RETURN list: " + list); }
        return list;
    }
    @Transactional
    public void removeChild(EntityPath path, ManagedObject child){
        ManagedObject[] pathWithChild = new ManagedObject[path.getSize(false) + 1];
        for (int i = 0; i < path.getSize(false); i++){
            pathWithChild[i] = path.getEntity(i);
        }
        pathWithChild[pathWithChild.length - 1] = child;
        
        PathProxy parentPp = this.getPathProxyFromMoArray(pathWithChild, false);
        
        this.getJpaTemplate().remove(parentPp);
    }
    
    /**
     * Takes a path of objects and returns the pathProxy for it, creating one if one doesn't exist already according to createIfNotFound.
     * @param voArray
     * @return
     */
    public PathProxy getPathProxyFromMoArray(ManagedObject[] moArray, boolean createIfNotFound){
        if (log.isInfoEnabled()) { log.info("= = = = = = =BEGIN getPathProxyFromVoArray(): " + ArrayUtils.toString(moArray)); }

        if (moArray == null || ArrayUtils.isEmpty(moArray)){
            if (log.isInfoEnabled()) { log.info("Got an empty voArray, returning null."); }
            return null;
        }

        // Build a query string that has as many name/id parameters as there are vo in the array
        // Note, this may not work with entities using composite keys - hibernate issues invalid sql for the null check at the end
        StringBuffer queryString = new StringBuffer("select pp from PathProxy pp "
            + "where ");

        if (moArray.length > 0) {
            for (int i = moArray.length - 1; i >= 0; i--){
                if (log.isTraceEnabled()) { log.trace("Creating ppClause for VO: " + i + " : " + moArray[i]); }

                StringBuffer ppClause = new StringBuffer("pp");
                // Add as many .parent's to the pp as we have counted into the hierarchy, this gets us the parent pp (eg, pp.parent.parent when i = 2)
                for (int parentCount = moArray.length - 2; parentCount >= i; parentCount--){ 
                    ppClause.append(".parent");
                }

                if ((moArray[i] != null) && (!moArray[i].equals("null"))){ // If not null, add type and id
                    String ppEntityType = ppClause.toString() + "." + PathProxy.ENTITY_TYPE_PROPERTY + " = '" + moArray[i].getClass().getName() + "'";
                    // Here's where we benefit from using the ManagedObject base class ... no need to figure out what property to use for the ID
                    String ppEntityId = ppClause.toString() + "." + PathProxy.ENTITY_ID_PROPERTY + " = " + moArray[i].getId();

                    queryString.append(ppEntityId);
                    queryString.append(" and ");
                    queryString.append(ppEntityType);

                    if (i > 0){ // Add an 'and' if we will add more ppClauses to the queryString
                        queryString.append(" and ");
                    }

                } else { // This vo is null, so add tp.parent = null
                    // This is where Hibernate generates bad sql for composite keys, removing fixes that:
                    // If you are using composite keys and getting an error, comment out the queryString.append line and expose the queryString.delete line
                    queryString.append(ppClause.toString() + " is null");
                    // Remove and added by last iteration (if you disable the 'is null')
//                    queryString.delete(queryString.length() - 4, queryString.length() ); //(" and ")
                }
            }
            
        } else {
            // Shouldn't get here
            return null;
        }

        if (log.isTraceEnabled()) { log.trace("------ Here's the queryString: " + queryString + " ------"); }

        List lst = this.getJpaTemplate().find(queryString.toString());
        if (log.isTraceEnabled()) { log.trace("Query for path proxy with moArray returned: " + lst); }

        PathProxy ppToReturn = null;

        if (CollectionUtils.isEmpty(lst)) { // No path proxy there yet
            if (createIfNotFound){
                if (log.isInfoEnabled()) { log.info("No pathproxy on that path, creating new..."); }
                ppToReturn = createPathProxy(moArray);
            } else {
                ppToReturn = null;
            }
        } else {
            ppToReturn = (PathProxy)lst.get(0);
        }
        if (log.isTraceEnabled()) { log.trace("RETURN ppToReturn: " + ppToReturn); }
        return ppToReturn;
    }
    /**
     * Creates a new TreePoint for the passed in vo array (voPath).  This method will walk the path up, getting tree points at every node, creating
     * them if necessary.
     * @param voPath
     * @return
     */
    @Transactional
    protected PathProxy createPathProxy(ManagedObject[] moPath){
        if (log.isInfoEnabled()) { log.info("| | | | | BEGIN createPathProxy: " + ArrayUtils.toString(moPath)); }
        // Get PP object
        PathProxy newPp = new PathProxy();
        
        int childIndex = moPath.length - 1;
        
        newPp.setEntityId(moPath[childIndex].getId());
        newPp.setEntityType(moPath[childIndex].getClass().getName());
        
        if (log.isTraceEnabled()) { log.trace("Set id: " + newPp.getEntityId() + " and type: " + newPp.getEntityType() + " on new PathProxy: " + newPp); }

        this.getJpaTemplate().persist(newPp);
        if (log.isTraceEnabled()) { log.trace("Saved newPp: " + newPp); }
        
        // Here is where we recursively build up the path until we find a parent path proxy that exists
        // Remove the current mo from path, to get the parent path
        Object[] objParentPath = ArrayUtils.remove(moPath, childIndex);
        // Convert back to mo array
        ManagedObject[] parentPath = new ManagedObject[objParentPath.length];
        for (int i = objParentPath.length - 1; i >= 0; i--){
            parentPath[i] = (ManagedObject)objParentPath[i];
        }
        
        // Get parent TreePoint
        PathProxy parentPp = (PathProxy)getPathProxyFromMoArray(parentPath, true);
        if (log.isTraceEnabled()) { log.trace("Here's the parentPp: " + parentPp); }
        // Set the parent
        newPp.setParent(parentPp);
        // This will either get the existing parentPp with its parents in place, or create a new one and get its parent, and so forth
        if (log.isTraceEnabled()) { log.trace("RETURN newPp: " + newPp + " with parent: " + newPp.getParent()); }
        
        return newPp;
    }

}
