package tutorial.util;

import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import sun.net.www.HeaderParser;
import tutorial.model.ManagedObject;
import tutorial.model.Project;
import tutorial.service.CommonService;

/**
 * <b>Created:</b> Apr 25, 2008<br>
 * <b>Title:</b> NodePath<br>
 * <b>Description:</b><br>
 * <b>Copyright:</b> Copyright (c) 2008<br>
 * @author MTyson
 * <p>
 * A simple class that models a path of entities in the system.
 * <br>
 * Note this object has the CommonService member injected by spring.  It is used by setting pathAsString.  In a real-world
 * system, this class would encapsulate various methods like 'isPathContainingType'.
 * <br>
 * This class can be passed between the front and back end, to hide how paths are represented and avoid any coupling.
 * <br>
 * This implementation represents string paths with the form PARENT_TOKEN^CHILD_TOKEN^GRANDCHILD_TOKEN, that is,
 * carrot-separated, from root to leaf.
 */

public class EntityPath {
    private static final transient Log log = LogFactory.getLog(tutorial.util.EntityPath.class);
    private CommonService commonService;
    
    private String pathAsString;
    private ManagedObject[] pathAsObjects;
    private ManagedObject[] pathAsObjectsNoHeaders;
    
    public static final String NODE_STRING_SEPARATOR = "^";
    
    /**default empty constructor*/
    public EntityPath() {
    }
    
    public EntityPath addChild(ManagedObject newChild){
        List<ManagedObject> newPath = new ArrayList<ManagedObject>();
        List<ManagedObject> newPathNoHeaders = new ArrayList<ManagedObject>();
        
        for (ManagedObject mo : this.getPathAsObjects()){
            newPath.add(mo);
        }
        
        for (ManagedObject mo : this.getPathAsObjectsNoHeaders()){
            newPathNoHeaders.add(mo);
        }
        
        newPath.add(newChild);
        
        if (!(newChild instanceof PathHeader)){
            newPathNoHeaders.add(newChild);
        }
        
        this.pathAsObjects = newPath.toArray(new ManagedObject[this.pathAsObjects.length + 1]);
        
        this.pathAsObjectsNoHeaders = newPathNoHeaders.toArray(new ManagedObject[this.pathAsObjectsNoHeaders.length + 1]);
        
        this.pathAsString = this.pathAsString + this.NODE_STRING_SEPARATOR + newChild.getClass().getName() 
            + this.NODE_STRING_SEPARATOR + newChild.getId();
        
        return this;
    }
    
    public EntityPath(CommonService commonService, String pathAsString){
        this.commonService = commonService;
        this.pathAsString = pathAsString;
    }
    
    public EntityPath setPathAsObjects(ManagedObject[] pathAsObjects) {
        this.pathAsObjects = pathAsObjects;
        return this;
    }

    public ManagedObject[] getPathAsObjects() {
        return getPathAsObjects(false);
    }
    public ManagedObject[] getPathAsObjects(boolean includeHeaders) {
        if (log.isInfoEnabled()) { log.info("BEGIN getPathAsObjects(): " + includeHeaders); }
        if (this.pathAsObjects != null){
        } else {
            if (StringUtils.isBlank(pathAsString)){
                throw new IllegalArgumentException("pathAsString must be set to get pathAsObjects");
            } else {
                this.pathAsObjects = this.getObjectPath(this.pathAsString);
            }
        }
        
        ManagedObject[] pathToReturn;
        
        if (includeHeaders){
            pathToReturn = this.pathAsObjects;
        } else {
            if (this.pathAsObjectsNoHeaders == null){
                List<ManagedObject> nonHeaders = new ArrayList(this.pathAsObjects.length);
                for (ManagedObject o : pathAsObjects){
                    if (!(o instanceof PathHeader)){
                        nonHeaders.add(o);
                    }
                }
                pathAsObjectsNoHeaders = nonHeaders.toArray(new ManagedObject[0]);
            }
            pathToReturn = pathAsObjectsNoHeaders;
        }
        if (log.isTraceEnabled()) { log.trace("RETURN pathToReturn: " + ArrayUtils.toString(pathToReturn)); }
        return pathToReturn;
    }
    
    public ManagedObject getEntity(int index){
        return this.getPathAsObjects()[index];
    }
    
    public int getSize(boolean includeHeaders){
        return this.getPathAsObjects(includeHeaders).length;
    }

    public String getPathAsString() {
        return pathAsString;
    }

    public EntityPath setPathAsString(String pathAsString) {
        this.pathAsString = pathAsString;
        return this;
    }
    
    /**
     * Returns an array of objects on a path.
     * <br>
     * This method requires a call to the back-end to populate the domain objects.  No DB activity is required until
     * this method is called.  The result is cached once that takes place the first time.
     * @param stringPath
     * @return
     */
    private ManagedObject[] getObjectPath(String nodeString){
        if (log.isInfoEnabled()) { log.info("BEGIN getObjectPath(): " + ArrayUtils.toString(nodeString)); }
        
        String[] stringPath = nodeString.split("\\"+NODE_STRING_SEPARATOR);
        
        List<ManagedObject> objects = new ArrayList<ManagedObject>();
        List<ManagedObject> nonHeaders = new ArrayList<ManagedObject>();
        
        Class type = null;
        Long id = null;
        for (int i = 0; i < stringPath.length - 1; i += 2){ // Count by 2s, since the path has type^id^type^id
            try {
                type = Class.forName(stringPath[i]);
            } catch (Exception e) {
                throw new RuntimeException("Unknown type: " + type);
            }
            
            if (type.equals(PathHeader.class)){
                objects.add(new PathHeader(stringPath[i + 1])); // In the case of headers, the second token is the label
            } else {
                id = Long.valueOf(stringPath[i+1]);
                objects.add((ManagedObject)commonService.get(type, id));
            }
        }
        return objects.toArray(new ManagedObject[stringPath.length / 2]);
    }
    
    public static String getObjectAsNodeString(ManagedObject mo){
        if (mo instanceof PathHeader){
            return mo.getClass().getName() + NODE_STRING_SEPARATOR + mo.getLabel();
        } else {
            return mo.getClass().getName() + NODE_STRING_SEPARATOR + mo.getId();
        }
    }

    public CommonService getCommonService() {
        return commonService;
    }

    public void setCommonService(CommonService commonService) {
        this.commonService = commonService;
    }
    
    public Object getFirstProjectParent(){
        for (int i = this.getPathAsObjects().length - 1; i >= 0; i--){
            if (this.getPathAsObjects()[i] instanceof Project){
                return this.getPathAsObjects()[i];
            }
        }
        return null;
    }
    
    public String toString(){
        return this.getPathAsString();
    }

    public boolean equals(Object o){
        if (log.isInfoEnabled()) { log.info("BEGIN equals: " + o + " | pathAsString: " + this.pathAsString); }
        if (o != null && this.getPathAsString() != null){
            return this.getPathAsString().equals(o.toString());
        }
        if (o == null && this.getPathAsString() == null){
            return true;
        }
        return false;
    }

    public ManagedObject[] getPathAsObjectsNoHeaders() {
        return pathAsObjectsNoHeaders;
    }
}
