package tutorial.webapp.backingbean;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import tutorial.jsf.component.ajaxtree.model.AjaxTreeDataProvider;
import tutorial.jsf.component.ajaxtree.model.AjaxTreeNode;
import tutorial.model.Developer;
import tutorial.model.ManagedObject;
import tutorial.model.Manager;
import tutorial.model.Person;
import tutorial.model.Project;
import tutorial.model.Task;
import tutorial.service.CommonService;
import tutorial.service.PathProxyService;
import tutorial.util.Constant;
import tutorial.util.EntityPath;
import tutorial.util.PathHeader;

/**
 * <b>Created:</b> Feb 4, 2008<br>
 * <b>Title:</b> AjaxTreeBean<br>
 * <b>Description:</b><br>
 * <b>Copyright:</b> Copyright (c) 2008<br>
 * @author MTyson
 * <p>
 * A backing bean to supply data to the Ajax Tree.
 */

public class AjaxTreeBean implements AjaxTreeDataProvider {
    private static final transient Log log = LogFactory.getLog(tutorial.webapp.backingbean.AjaxTreeBean.class);

    private static List<AjaxTreeNode> rootNodes;

    private CommonService commonService;
    private PathProxyService pathProxyService;
    
    public AjaxTreeBean() {
    }
    
    /**
     * Here's where it all starts.  This method defines what nodes are at the root, and what data they hold.
     * It simply returns a list of type AjaxTreeNode.  The most important thing here is the argument of the 
     * AjaxTreeNode constructor, the 'nodeString'.  That is what will identify the node to us when the user clicks on it
     * and it is being expanded.  We'll have the nodeString when we execute the getChildNodes method.
     */
    public List<AjaxTreeNode> getRootNodes(){
        rootNodes = new ArrayList<AjaxTreeNode>();
        rootNodes.add(new AjaxTreeNode(Constant.PEOPLE, false, "People", Constant.PEOPLE));
        Long projectCount = commonService.getCount(Project.class);
        rootNodes.add(new AjaxTreeNode(Project.class.getName(), projectCount <= 0, "Projects", Constant.PROJECTS));
        return rootNodes;
    }
   
    /**
     * Everytime a node is clicked in the tree, it sends back the nodeString defined for it to this method.  The job here
     * is to return an array of type AjaxTreeNode representing the nodeString's children.
     */
    public List getChildNodes(String nodeString){
        if (log.isInfoEnabled()) { log.info("BEGIN getChildNodes(): " + nodeString); }
        
        EntityPath path = this.getNewPath().setPathAsString(nodeString); // Get an EntityPath object representing the path
        
        ArrayList<AjaxTreeNode> children = new ArrayList<AjaxTreeNode>();
        if (path.equals(Constant.PEOPLE)){
            Long devCount = commonService.getCount(Developer.class);
            children.add(new AjaxTreeNode(Developer.class.getName(), devCount <= 0, "Developers"));
            
            Long managerCount = commonService.getCount(Manager.class);
            children.add(new AjaxTreeNode(Manager.class.getName(), managerCount <= 0, "Managers", "normal", "managers"));
        } else if (path.equals(Developer.class.getName()) 
                || path.equals(Manager.class.getName()) || path.equals(Project.class.getName())){
            try {
                children.addAll(this.getMoNodeList(commonService.getAll(Class.forName(path.toString()))));
            } catch (Exception e){
                throw new RuntimeException("Unknown type (classname expected): " + path);
            }
        } else if (path.getSize(true) > 0){ 

            Project clickedProject = (Project)path.getEntity(0); // Get the project, which is the first object in the path
            
            // In this system, only a project instance is alone in a path
            if (path.getSize(true) == 1) { // A project instance 
                if (path.getPathAsString().indexOf("managersOnProject") == -1) {
                    if (log.isTraceEnabled()) { log.trace("clickedProject: " + clickedProject); }
                    int manOnProjectCount = clickedProject.getManagers().size(); 
                    if (log.isTraceEnabled()) { log.trace("manOnProject: " + manOnProjectCount); }
                    
                    PathHeader header = new PathHeader("Managers On Project");
                    
                    children.add(new AjaxTreeNode(nodeString + EntityPath.NODE_STRING_SEPARATOR + EntityPath.getObjectAsNodeString(header), 
                            manOnProjectCount <= 0, header.getLabel()));
                }
            } else if (path.getSize(true) == 2) { 
                if (log.isTraceEnabled()) { log.trace("Opening managersOnProject node."); }
                for (Manager manager : clickedProject.getManagers()){
                    String manOnProjectNodeString = nodeString + EntityPath.NODE_STRING_SEPARATOR + EntityPath.getObjectAsNodeString(manager);
                    // Notice we just add the new child node (manager) to the path to get the child developers, to determine the leafs status of the node
                    List<Object> projectManagerDev = pathProxyService.getPathChildren(path.addChild(manager), Developer.class.getName());
                    children.add(new AjaxTreeNode(manOnProjectNodeString, projectManagerDev.size() <= 0,
                            manager.getFirstName(), Manager.class.getName(), manager));
                }
                
            } else if (path.getSize(true) == 3) { // project->header->manager
                List<Object> projectManagerDev = pathProxyService.getPathChildren(path, Developer.class.getName());
                for (Object dev : projectManagerDev){
                    Developer developer = (Developer)dev;
                    String devNodeString = nodeString + EntityPath.NODE_STRING_SEPARATOR + EntityPath.getObjectAsNodeString(developer);
                    List<Object> projectManagerDevTask = pathProxyService.getPathChildren(path.addChild(developer), Task.class.getName());
                    children.add(new AjaxTreeNode(devNodeString, projectManagerDevTask.size() <= 0,
                            developer.getFirstName(), Developer.class.getName(), developer));
                }
            } else if (path.getSize(true) == 4){ // Again, we know a three node path is project->manager->developer because the system is so simple
                List<Object> tasksOnDev = pathProxyService.getPathChildren(path, Task.class.getName());
                children.addAll(this.getMoNodeList(tasksOnDev));
            }
        } else {
            throw new IllegalArgumentException("Unknown node type: " + nodeString);
        }
        if (log.isInfoEnabled()) { log.info("RETURNING children: " + children); }
        return children;
    }
    
    /**
     * Creates a list of AjaxTreeNodes for simple cases
     * @param entities
     * @return
     */
    protected List<AjaxTreeNode> getMoNodeList(List<Object> entities){
        List<AjaxTreeNode> nodes = new ArrayList<AjaxTreeNode>();
        for (Iterator<Object> iter = entities.iterator(); iter.hasNext();){
            ManagedObject mo = (ManagedObject)iter.next();
            nodes.add(new AjaxTreeNode(mo.getClass().getName() + EntityPath.NODE_STRING_SEPARATOR + mo.getId(), 
                    !(mo instanceof Project), mo.getLabel(), mo.getClass().getName(), mo)); // Quick and dirty - we know Project has children (the headers), and assume that dev, manager + task don't (this method is only used for root-level objects)
        }
        return nodes;
    }

    public CommonService getCommonService() {
        return commonService;
    }

    public void setCommonService(CommonService commonService) {
        this.commonService = commonService;
    }

    public PathProxyService getPathProxyService() {
        return pathProxyService;
    }

    public void setPathProxyService(PathProxyService pathProxyService) {
        this.pathProxyService = pathProxyService;
    }
    /**
     * An infrastructure method.  This is a 'method-lookup' injection from Spring.  Spring will inject a new 
     * entityPath bean here (prototype scoping).  @see applicationContext.xml
     * @return
     */
    public EntityPath getNewPath(){
        // Spring injected
        return null;
    }
}
