/**
 * JavaScript file for the AjaxTree.  The two main classes are AjaxTree and TreeNode.  The AjaxTree class maintains
 * a static array of trees in the document, making the component safe for multiple instances.
 * <p>
 * TODO: May need to protect the TreeNode classname in a namespace.  Maybe add it to AjaxTree.
 * <p>
 * Currently there are two custom HTML attributes used to send data.  These are non-standard html.  They are
 * the nodeString and leaf attributes, which are set on the root nodes and sent with every child over Ajax.
 * These could be sent as hidden fields instead.  Will add the ability to select either way (performance or
 * compliance).
 * <p>
 * @author - Matt Tyson
 * Begun: Feb. 4, 2008
 */

/**
 * Class to model the ajaxTree.  
 * Includes static method loadTrees to initialize all existing trees.
 * <p>
 * 
 * Developed with the aid of FireFox FireBug.  Give it a try.
 * @param - domNode: The outer div wrapper for the tree (class = AjaxTree.TREE_CLASS)
 */
function AjaxTree(/*domNode*/ domNode) {
	this.domNode = domNode; // A 'member' variable to hold the domNode for this tree
	
	// Attach event handler to the tree.  Note, we catch events on the tree instead of attaching listeners to all the nodes
	dojo.event.connect(domNode, "onclick", this, "treeClicked"); // Note we call the method on the tree instance
	
	// An assoc array to hold the tree's nodes
	// keyed on ID, value is a TreeNode object
	this.nodes = new Array();
	
	// Get the child nodes of the tree, so we can get the elements we are interested in
	var children = domNode.childNodes;
	
	// We get all the roots present.  We need to know which is the last one.
	var roots = dojo.html.getElementsByClass(AjaxTree.ROOT_INPUT_CLASS, domNode, "input"); 
	
	for (var i = 0; i < roots.length; i++){
		var rootData = roots[i].value;
		var rootNodeData = eval("("+rootData+")"); // Evaluate the JSON object
		var lastChild = (i == (roots.length - 1));  // Is it the last root node?
		rootNodeData.isLastChild = lastChild;
		rootNodeData.nestingLevel = 0;
		rootNodeData.ajaxTree = this;
		rootNodeData = TreeNode.createDomNode(rootNodeData); // Generate the dom for this node
		new TreeNode(rootNodeData);
	}
	
	// Save the server url, which is in a hidden field, in a member var
	var urlHidden = dojo.html.getElementsByClass(AjaxTree.SERVER_URL_CLASS, domNode, "input"); 
	this.serverUrl = urlHidden[0].value; // Should be only one of these guys (per tree)
}

/**
 * @return A nodeData json object with the data for the treeNode dom passed in
 */
AjaxTree.parseTreeNode = function(/*domNode*/domNode){
	var nodeData = {};
	
	nodeData.treeNodeDiv = domNode;
	
	// Iterate over the nodes within the treeNode, looking for elements we are interested in
	var nodeChildren = domNode.childNodes; // These are dom children
	for (var i in nodeChildren){
		if (nodeChildren[i].name == AjaxTree.NODE_STRING_NAME) { // get the nodestring hidden field
			// NOTE: Some users may want to avoid custom html attributes, and in that case we could use hidden
			// fields for nodeString and leaf status.  Here we are looking for the nodeString hiddne field
			// Right now, we are using a custom attribute. 
			// Found the hidden nodeString field
			var nodeStringHidden = nodeChildren[i];
			nodeData['nodeString'] = nodeStringHidden.value; // Set the nodeString
		} else if (nodeChildren[i].className == AjaxTree.NODE_LABEL_CLASS){ // nodeLabel
			nodeData.nodeLabel = nodeChildren[i];
			// This is the nodeLabel div, which contains the clickArea and userData divs
			var labelChildren = nodeData.nodeLabel.childNodes;
			for (var labelCount in labelChildren){
				if (labelChildren[labelCount].className == AjaxTree.TREE_DATA_CLASS) {
					nodeData.treeDataDiv = labelChildren[labelCount];
					var treeDataChildren = nodeData.treeDataDiv.childNodes;
					for (var treeDataCount = 0; treeDataCount < treeDataChildren.length; treeDataCount++){
						if (TreeNode.classContainsToken(treeDataChildren[treeDataCount], AjaxTree.CLICK_AREA_CLASS)){
							// the clickArea
							nodeData.clickArea = treeDataChildren[treeDataCount];
						} else if (TreeNode.classContainsToken(treeDataChildren[treeDataCount], AjaxTree.FILL_DIV_CLASS)){
							// Fillers
							nodeData.fillDivs = new Array();
							nodeData.fillDivs[nodeData.fillDivs.length] = treeDataChildren[treeDataCount];
						}
					}
				} else if (labelChildren[labelCount].className == AjaxTree.USER_DATA_CLASS){
					// the userData
					nodeData.userDataDiv = labelChildren[labelCount];
				}
			} //labelChildren
		} else if (nodeChildren[i].className == AjaxTree.CHILD_WRAPPER_CLASS) { // childWrapper
			nodeData.childWrapper = nodeChildren[i];
		}
	} // nodeChildren
	
	// Get the nodeString from custom html attr.
	nodeData.nodeString = domNode.getAttribute(AjaxTree.NODE_STRING_NAME);
	
	nodeData.expanded = false; // The state of the node
	nodeData.isLeaf = domNode.getAttribute(AjaxTree.LEAF_STATUS_NAME) == "true" ? true : false; // Currently, we are only concerned about whether there are 0 or 0> children, but in the future, the actual child count could be used
	nodeData.childrenLoaded = false; // Tree node begins life without its child data loaded
	
	nodeData.childCache = new Array(); // An array that holds all this child's children
	
	return nodeData;
}

/**
 * 'static' method to initialize all the trees on the page.
 */
AjaxTree.loadTrees = function(){
//	AjaxTree.bodyDebug("Loading");
	
	// Get all the trees on document (ie, all the tree 'wrapper' divs)
	var trees = dojo.html.getElementsByClass(AjaxTree.TREE_CLASS, null, "div");
	
	// Create a new tree object for each tree found
	for (var i in trees){
		var newTree = new AjaxTree(trees[i]); // Create a tree and set domNode
		AjaxTree.allTrees[trees[i].id] = newTree; // Keyed on dom id
	}
}

AjaxTree.prototype.treeClicked = function(evt){
	// The tree has been clicked, now we need to find out what was clicked
	var target = evt.target;

	// We only care here if a clickArea has been clicked
	if (target.className && TreeNode.classContainsToken(target, AjaxTree.CLICK_AREA_CLASS)){ // Does the class contain clickArea?
		// Now we need to get the treeNode div.  First, get the nodeLabel div that wraps the clickArea
		var nodeLabel = target.parentNode; // label wraps clickArea
		var treeNode = AjaxTree.getContainerByClass(nodeLabel, AjaxTree.TREE_NODE_CLASS);  // treeNode wraps label

		// Question: We know the id contains the AjaxTree's dom id, is it faster to use document.getElementById
		//   or start at the node and crawl the parent's back to the AjaxTree
		// Decision: Going with crawling parents to avoid parsing the id string
		// Get the tree dom node that was clicked
		var ajaxTreeDom = AjaxTree.getContainerByClass(treeNode, AjaxTree.TREE_CLASS); // Get the tree containing this node

		// Out of all the trees, get our tree
		var ajaxTreeObject = AjaxTree.allTrees[ajaxTreeDom.id];

		var treeNodeObject = ajaxTreeObject.nodes[treeNode.id]; // Now we get the treeNode object
		
		if (!treeNodeObject.isLeaf){ // Leaves don't expand or collapse
			treeNodeObject.changeState(); // Change the state
			treeNodeObject.render(); // State has changed, so render the node
		}
	}
}

/**
 * Constructor for the TreeNode class.
 * <p>
 * domNode.id must be unique inside the tree.  Considered using dojo.dom.getUniqueId.  Decided on using a 
 * path for the id that is built from parent.id and a unique int.  This avoids the dom parsing of dojo.dom.getUniqueId
 * <p>
 * Sets the necessary member variables and parses out the internal structure.
 * <p>
 * Either the domNode will have been rendered out by the server in the case of a root, or it will have been
 * generated on the client for a non-root node sent over Ajax.  In either case, the domNode should 
 * be inserted into the document as a child of its parent, since this constructor doesn't do it.
 * @param domNode - should be the div with class = AjaxTree.TREE_NODE_CLASS, if null, nodeData is required
 * @param tree -- The tree to which this node belongs (required)
 * @param parentNode -- The TreeNode object that is the parent (optional, won't be present for root)
 * @param nestLevel -- How many parents this node has.  If null, 0 is assumed (root)
 * @param nodeData -- an XML document containing the info required to construct the node.  If null, domNode is required.
 * <p>
 * nodeData:
 *  'isLeaf'       : kidIsLeaf,
	'isLastChild'  : lastChild,
	'nodeString'   : kidNodeString,
	'userDataHTML' : userDataHTML,
	'parentNode'   : parentNode,
	'uniqueId'     : kidId,
	'nestingLevel' : nestLevel,
	'ajaxTree'     : AjaxTree,
	'treeNodeDiv'  : domNode,
	'isExpanded'   : boolean,
	'treeDataDiv'  : dom node,
	'userDataDiv'  : dom node,
	'fillDivs'     : node array,
	'padDivs'      : node array,
	'labelDiv'     : dom node,
	'childWrapper' : dom node
 */
function TreeNode(/*json object*/ nodeData){		
	// Some of the nodeData pieces are used alot, so we add them to the node itself
	// TODO: Can we just eleminate the json object and use the TreeNode object...?
	this.nodeData = nodeData;
	this.domNode = nodeData.treeNodeDiv; // Access to the treeNode dom element
	this.childWrapper = nodeData.childWrapper;
	this.isExpanded = nodeData.isExpanded;
	this.parentNode = nodeData.parentNode;
	this.ajaxTree = nodeData.ajaxTree;
	this.nodeString = nodeData.nodeString;
	this.nestingLevel = nodeData.nestingLevel;
	this.isLastChild = nodeData.isLastChild;
	this.userData = nodeData.userDataDiv;
	this.treeData = nodeData.treeDataDiv;
	this.fillDivs = nodeData.fillDivs;
	this.clickArea = nodeData.clickArea;
	this.isLeaf = nodeData.isLeaf;
	this.labelDiv = nodeData.labelDiv;

	if (this.ajaxTree){
		// Add node to the tree's node cache
		this.ajaxTree.nodes[this.domNode.id] = this;
	} else {
		// Error state
		alert("Missing ajaxTree in nodeData.");
	}
	
	// If there is a parent, attach the domNode to its childWrapper
	if (this.parentNode){
		this.parentNode.childWrapper.appendChild(this.domNode);
	} else {
		// No parent, so append to the tree itself (its a root node)
		this.ajaxTree.domNode.appendChild(this.domNode);
	}
	
	// Now that the node is attached to the dom, we can get hieghts from things
	// We want the taller of the two, between the user and tree data divs
	this.dataHeight = this.userData.offsetHeight >= this.treeData.offsetHeight 
				? this.userData.offsetHeight 
				: this.treeData.offsetHeight;
	this.labelDiv.style.height = this.dataHeight; // make label div the same height
	
	// Make the treeData and userData divs are the same height
	this.treeData.style.height = this.dataHeight; 
	this.userData.style.height = this.dataHeight;
	
	// The fill height is the remainder of the treeData area after the clickArea/padDivs
	this.fillHeight = (this.userData.offsetHeight - nodeData.clickArea.offsetHeight); // The filler height is the rest of tree data, after pad/clickArea
	
	// Refresh the fillDivs
	this.refreshFillDivs(true); // Set the height also
}

TreeNode.classContainsToken = function(/*dom node*/ domNode, /*string*/token){
	if (domNode && domNode.className){
		return domNode.className.indexOf(token) > -1;
	}
	return false;
}

/**
 * Sets the display value on all the fillDivs.
 * @param setHeight - If true, will calculate and set the height for the fillDivs (Usually done only when the node is first created)
 */
TreeNode.prototype.refreshFillDivs = function(/*boolean*/setHeight){
	for (var i = 0; i < this.fillDivs.length; i++){
		if (setHeight){
			this.fillDivs[i].style.height = this.fillHeight;
		}
		if (!this.isExpanded) {
			// hide it (filler is only visible when there are children visible, or there are siblings below)
			// Slightly tricky/hacky: If the fill div is under the clickArea (ie, the last filler) and the node is closed, don't display
			if (this.fillDivs.length - 1 == i){
				this.fillDivs[i].style.display = "none";
			}
		} else {
			this.fillDivs[i].style.display = "block";
		}
	}
}

/**
 * This method is called when a node's clickArea has been clicked, initiating an expansion of collapse.
 * Notice that all we do is set the exanded state on the node.  TreeNode.render is called to handle the drawing
 * of the node.
 */
TreeNode.prototype.changeState = function(){
//	alert("My state is changing.");
	// The node's state is changing, the first thing we do is check what the previous state was, and change it
	// along with the class of the clickArea (this changes the expansion icon)
	if (this.isExpanded){ // Node was expanded, collapse it
		this.isExpanded = false;
	} else { // Node was closed, open it
		this.isExpanded = true;
	}
}

TreeNode.prototype.getClickAreaClass = function(){
	var className = AjaxTree.CLICK_AREA_CLASS;
	if (this.nestingLevel == 0){ // root
		if (this.isLeaf) { // root w/no kids
			className += " " + AjaxTree.ROOT_LEAF_CLASS;
		} else { // Root with kids
			if (this.isExpanded) { // Open root
				className += " " + AjaxTree.ROOT_OPEN_CLASS;
			} else { // Closed root
				className += " " + AjaxTree.ROOT_CLOSED_CLASS;
			}
		}
	} else { // Non-root
		if (this.isLeaf) { // non-root w/no kids
			className += " " + AjaxTree.NODE_LEAF_CLASS;
		} else { // Root with kids
			if (this.isExpanded) { // Open root
				className += " " + AjaxTree.NODE_OPEN_CLASS;
			} else { // Closed root
				className += " " + AjaxTree.NODE_CLOSED_CLASS;
			}
		}
	}
	return className;
}
/**
 * Draws the node.  If the node is expanded and the child data hasn't been loaded, this method will initiate
 * an Ajax call to fetch the data.
 */
TreeNode.prototype.render = function(){
//	AjaxTree.bodyDebug("render");
	if (this.isExpanded){ 
		// Check if we have the server data to render the node
		if (!this.childrenLoaded){ // We don't have the children yet
			var urlToPost = this.ajaxTree.serverUrl;
			
			this.loadingDataMsg(true); // Add the loading data msg
			this.requestChildData(urlToPost);
		}
		// Make the children visible (by removing the display style)
		this.nodeData.childWrapper.style.display = "block";
	} else { // Node is collapsed
		// Make the children NOT visible 
		this.nodeData.childWrapper.style.display = "none";
	}
	this.clickArea.className = this.getClickAreaClass(); // Reset the clickArea className, so its icon is correct
	// Refresh the fill divs
	this.refreshFillDivs(false); // Don't need to set height
}

/**
 * Takes care of the server's response to a request for child data.
 */
TreeNode.prototype.handleChildDataResponse = function(/**JSON response*/response){
	// Remove the loading data msg
	this.loadingDataMsg(false);
	
//	var responseJSON = eval("({kids : {0 : { isLeaf : true, userDataHTML : 'test'}, 1 : { isLeaf : false} }, stat : 'OK' })");
	var responseJSON = eval("(" + response + ")");
	
//	AjaxTree.bodyDebug("stat: " + stat);
	if (responseJSON.stat == "ERROR"){
		alert("Failed to retrieve data.");
		// TODO: Allow user to determine whether to raise alert or just clean up tree.
	} else if (responseJSON.stat == "OK"){ // Successfully retrieved child data
		// Now we've got to create dom nodes for each child

//		AjaxTree.bodyDebug("kids: " + responseJSON.kids.length);
		// TODO: I think using JSON will be simpler than xml
//		for (var kidIter = 0; kidIter < responseJSON.kids.length; kidIter++){
		for (var kidIter = 0; kidIter < responseJSON.kids.length; kidIter++){
			// A couple bits of info we will add to the child json
			var lastChild = kidIter == (responseJSON.kids.length - 1); //Is this the last child of the parent?
			// Create the ID
			var kidId = TreeNode.createNodeId(this.domNode.id, kidIter);
		
			// Marshall all the info needed to create a treeNode
			responseJSON.kids[kidIter].isLastChild = lastChild;
			responseJSON.kids[kidIter].parentNode = this;
			responseJSON.kids[kidIter].uniqueId = kidId;
			responseJSON.kids[kidIter].ajaxTree = this.ajaxTree;
			responseJSON.kids[kidIter].nestingLevel = (this.nestingLevel + 1);
		
			// Create the new domNode, which returns a json object full of info
			kidNodeData = TreeNode.createDomNode(responseJSON.kids[kidIter]); // Create the new domNode

			// Instantiate the node
			var kidTreeNode = new TreeNode(responseJSON.kids[kidIter]);
		}
	}
	
	this.childrenLoaded = true; // Flag the node as being loaded
}

TreeNode.createNodeId = function(/*string*/parentId, /*int*/uniqueInt){
	return parentId + ":" + uniqueInt;
}
/**
 * Creates a fully functional domNode for the passed in data. 
 * <p>
 * We keep track of the node parts in the nodeData json object
 * @return A json object that includes the data passed in, as well as the parts of the new domNode
 */
TreeNode.createDomNode = function(/*json*/nodeData){
	// Create the node itself
	var treeNodeDiv = document.createElement("div");
	treeNodeDiv.className = AjaxTree.TREE_NODE_CLASS;
	treeNodeDiv.id = nodeData.uniqueId;
	nodeData['treeNodeDiv'] = treeNodeDiv; // Add the treeNode to the nodeData
	
	// USING CUSTOM HTML ATTR INSTEAD OF HIDDEN FIELD
	// Create a hidden input with the nodeString
	// This keeps things the same as the root node, but does mean extra dom processing.  We could pass the nodeString
	// into the TreeNode constructor instead.
//	var nodeStringHidden = document.createElement("input");
//	nodeStringHidden.type = "hidden";
//	nodeStringHidden.className = AjaxTree.NODE_STRING_NAME;
//	nodeStringHidden.value = nodeString;
////	nodeStringHidden.name = AjaxTree.NODE_STRING_NAME;
//	treeNodeDom.appendChild(nodeStringHidden); // Set the nodeString field on the node

	// Create the label
	var labelDiv = document.createElement("div");
	labelDiv.className = AjaxTree.NODE_LABEL_CLASS;
	nodeData['labelDiv'] = labelDiv;

	// The treeData block
	var treeData = document.createElement("div");
	treeData.className = "treeData";
	nodeData['treeDataDiv'] = treeData;
	
	//TODO: Make the fill div stuff optional
	// If a user just wants to contain all node data on one line, the fill divs aren't necessary
	// (Once the height of the nodes is greater)
	
	// Now insert the padding divs
	var padDivs = TreeNode.createPaddingDivs(nodeData.parentNode, nodeData.isLastChild); 
//	var padHeight = padDivs[0].style.height; // Get the height of the pad divs (we'll use it later to determine the fill div height)
	var fillDivs = new Array();
	// Now insert each pad div into the label div
	// NOTE: Extracting the fillDivs to its own method would be cleaner, but would be less performant ... premature optimization?
	for (var padCount = 0, fillCount = (padDivs.length - 1); 
					padCount < padDivs.length; 
					padCount++, fillCount--){
		treeData.appendChild(padDivs[padCount]);
		// We also create the filler div
		// We need a fill div for each pad, and one for the click area
		fillDivs[padCount] = document.createElement("div");
		fillDivs[padCount].className = AjaxTree.FILL_DIV_CLASS;
		if (TreeNode.doesPadDivHaveLineBelow(padDivs[padCount])){
//				&& !nodeData.lastChild) { // That is, the pad div has a line below it
			// So we want the fill div under it to contain a line also
			fillDivs[padCount].className += " " + AjaxTree.PAD_DIV_LINE_CLASS;
		}
	}
	// Now, add a last fill div, the one that will go under the clickArea
	fillDivs[fillDivs.length] = document.createElement("div");
	fillDivs[fillDivs.length - 1].className = AjaxTree.FILL_DIV_CLASS;
	if (!nodeData.isLeaf) { // If its not the last child, we fill in lines below click area (if its not a root)
		fillDivs[fillDivs.length - 1].className += " " + AjaxTree.PAD_DIV_LINE_CLASS;
	}
	nodeData['padDivs'] = padDivs;
	nodeData['fillDivs'] = fillDivs;	
	
	// Create clickArea
	var clickArea = TreeNode.createClickArea(nodeData.isLeaf);
	nodeData['clickArea'] = clickArea;
	
	// Append clickArea to the labelDiv
	treeData.appendChild(clickArea); 
	// Now the breaking div, the pushes the fill divs down
	var breakDiv = document.createElement("div");
	breakDiv.className = "break";
	treeData.appendChild(breakDiv);
	
	// We need the user div now, for the height
	// Create userData div
	var userDataDiv = document.createElement("div");
	userDataDiv.className = AjaxTree.USER_DATA_CLASS;
	// Stick the new html into the label
	userDataDiv.innerHTML = nodeData.userDataHTML;
//	var userHeight = userDataDiv.offsetHeight;
	nodeData['userDataDiv'] = userDataDiv;
	
	// Add the filler divs, using the userData height to determine how tall they need to be
	for (var x in fillDivs){
		treeData.appendChild(fillDivs[x]);
//		fillDivs[x].style.height = userHeight - padHeight;
	}
	
	labelDiv.appendChild(treeData);

	// Append userData to label
	labelDiv.appendChild(userDataDiv);
	
//	treeData.style.height = userHeight;

	// Append label to treeNode
	treeNodeDiv.appendChild(labelDiv);
	
	// Create and append the childWrapper
	var childWrapper = document.createElement("div");	
	childWrapper.className = AjaxTree.CHILD_WRAPPER_CLASS;
	treeNodeDiv.appendChild(childWrapper);
	nodeData['childWrapper'] = childWrapper;
	
	return nodeData;
}

TreeNode.doesPadDivHaveLineBelow = function(/*div*/padDiv){
	var className = padDiv.className;
	if (className.indexOf(AjaxTree.PAD_DIV_LAST_CHILD_CLASS) > -1){
		return false; // Last child pad does not have line below
	}
	if (className.indexOf(AjaxTree.PAD_DIV_MID_CHILD_CLASS) > -1){
		return true;  // Mid child does have line below
	}
	if (className.indexOf(AjaxTree.PAD_DIV_LINE_CLASS) > -1){
		return true;
	}
	if (className.length == AjaxTree.PAD_DIV_CLASS.length){
		return false; // A pad div with no line doesn't have line below 
	}
}

TreeNode.createClickArea = function(/*true|false*/isLeaf){
	// Create the clickArea
	var clickArea = document.createElement("div");
	clickArea.className = AjaxTree.CLICK_AREA_CLASS;
	// Now we will either give the click area an expandable icon or the leaf icon
	if (isLeaf){
		clickArea.className += " " + AjaxTree.NODE_LEAF_CLASS; // The no-expand icon
	} else {
		clickArea.className += " " + AjaxTree.NODE_CLOSED_CLASS; // The closed branch icon
	}
	return clickArea;
}

/**
 * The 'padding divs' is used to push the node out to the right (or left in the future) as far as needed.
 * They also display the tree lines from ancestor nodes.
 * @param parentNode (required) - Since the padding divs represent the relation between a node and its parents, we need the parent TreeNode
 * @param lastChild (required) - Is the node we are creating pads for the last child of parent? (Affects the image the pad will have)
 * @return - An array of divs
 */
TreeNode.createPaddingDivs = function(/**parentNode*/parentNode, /**boolean*/lastChild){
	var divs = new Array();
	if (parentNode) { // If no parent node, no padding divs are required
		var nestingLevel = parentNode.nestingLevel + 1; // The nesting level is the parent nesting level + 1
		// divCount is the index for the return array (we want it to go in reverse order, from 0 up, so the client of the method has the divs in order)
		for (var nestCount = nestingLevel, divCount = 0; 
					nestCount >= 0 && parentNode != null; 
					nestCount--, divCount++){
			// Note that parentNode.parentNode should become null at the SAME time that the nestingLevel
			// iterator (nestCount) breaks out of this loop.  But just in case we add the null check.
			var newPad = document.createElement("div");
			newPad.className = AjaxTree.PAD_DIV_CLASS; // Always gets the pad div class
			// Now add the pad class that determines the image		
			// If we are looking at the first parent, there will always be a line
			// Sometimes its an angle, sometimes a three way branch, depending on if there's more children or not
			if (nestCount == nestingLevel){ // Meaning, this is the first pad, so the pad under the first parent
				if (lastChild){ // Pad is last child under immediate parent, gets the angle line
					newPad.className += " " + AjaxTree.PAD_DIV_LAST_CHILD_CLASS;
				} else { // Pad of a child where more children will appear below, gets the three way branch
					newPad.className += " " + AjaxTree.PAD_DIV_MID_CHILD_CLASS;
				}
			} else { // NOT the first parent
				// This is a little tricky.  Now we need to know if the ancestor just below the one above this pad 
				// was the last child.  That is, does this pad need to show a line?  
				if (parentNode.isLastChild){
					// Do nothing, the padding div does not require an image
				} else { // The ancestor has other children, pad needs a line
					if (nestingLevel == 0){ // We are under the root level.  Depending on tree style, may want lines or spaces
						// Do nothing, don't attach roots with lines
					} else {
						newPad.className += " " + AjaxTree.PAD_DIV_LINE_CLASS;
					}
				}
				parentNode = parentNode.parentNode; // Climb up a node
			}		
			
			divs[nestCount - 1] = newPad; // Assign the index - 1, so we fill it up down to index = 0
		}
	}
	return divs;
}

/**
 * Makes an ajax call to the server to get the child data for this node.
 * <p>
 * TODO: If more request types are added, extract the actual ajax call to another method.
 */
TreeNode.prototype.requestChildData = function(/**string*/urlToPostOld){
	var urlToPost = this.ajaxTree.urlToPost;
	if (!urlToPost){
		var form = dojo.dom.getFirstAncestorByTag(this.ajaxTree.domNode, "form");
		if (!form){
			alert("AjaxTree must be contained in a form.");
		}
		urlToPost = form.action;
	}
	var params = {}
	params[AjaxTree.NODE_STRING_REQUEST_KEY] = this.nodeString; // Identify the node to the server
	params[AjaxTree.AJAX_REQUEST_KEY] = "true"; // Flag the request as Ajax
	params[AjaxTree.CLIENT_ID_REQUEST_KEY] = this.ajaxTree.domNode.id; // Identify the component to the ajaxListener
	params[AjaxTree.REQUEST_TYPE_REQUEST_KEY] = AjaxTree.REQUEST_TYPE_EXPAND; // We are asking for child data from expansion
	
	// For compatibility with JSF 1.2 - This field identifies the view
	var viewState = document.getElementById("javax.faces.ViewState");
	if (viewState){
		params["javax.faces.ViewState"] = viewState.value;
	}
	
	var treeNode = this; // Put the current treeNode in a var, so we can pass it into dojo.bind (passing 'this' will resolve to the dojo.bind method itself)
	
	dojo.io.bind({
				url: urlToPost,
				method: "POST",
				load: function(type, data, evt){ treeNode.handleChildDataResponse(data); },
				mimetype: "text/plain",
				content: params
			});
}

/**
 * Adds or removes the 'loading data...' message on the node.
 */
TreeNode.prototype.loadingDataMsg = function(/**boolean*/ display){
	if (display){
		var loadingNodeData = {
			'isLeaf'       : true,
			'isLastChild'  : true,
			'nodeString'   : 'LoadingData',
			'userDataHTML' : 'Loading Data ...',
			'parentNode'   : this,
			'uniqueId'     : this.domNode.id + ":LOADING_DATA",
			'nestingLevel' : this.nestingLevel + 1,
			'ajaxTree'     : this.ajaxTree,
			'isExpanded'   : false
		}
		var loadingDataDiv = TreeNode.createDomNode(loadingNodeData);
		var loadingTreeNode = new TreeNode(loadingDataDiv);
		this.childWrapper.appendChild(loadingTreeNode.domNode);
	} else {
		var wrapperKids = this.childWrapper.childNodes;
		for (var node in wrapperKids){
			if (wrapperKids[node].id == this.domNode.id + ":LOADING_DATA"){
				this.childWrapper.removeChild(wrapperKids[node]);
			}
		}
	}
}

/**
 * A convienience method which takes any domNode and crawls the parents until it finds a node with class = className
 * or null if not found.
 */
AjaxTree.getContainerByClass = function(/*dom Node*/ domNode, /*String*/className){
	if (domNode.className == className){
		return domNode; // The passed domNode is the tree
	} else {
		if (domNode.parentNode == null){
			return null; // Failed to find a tree
		} else {
			return AjaxTree.getContainerByClass(domNode.parentNode, className); // Recurse on parent
		}
	}
}

AjaxTree.bodyDebug = function(msg){
	var para = document.createElement("p");
	var text = document.createTextNode(msg);
	para.style.fontSize = 11;
	para.appendChild(text);
	document.body.appendChild(para);
}

AjaxTree.debugObject = function(obj){
	AjaxTree.bodyDebug("obj: " + obj);
	for (x in obj){
		AjaxTree.bodyDebug("member: " + x + " = " + obj.x);
	}
}

/**
 * The master list of all the trees in the document.  This makes the component safe for multiple instances.
 * <p>
 * Associative array from dom id -> AjaxTree object
 */
AjaxTree.allTrees = new Array(); 

// These constants should be safe because they are only searched on from within a tree or a node
// However, if they are ever required to be searched on in the whole document, they should be namespaced (As the tree class is)
AjaxTree.TREE_CLASS = "tutorial.AjaxTree";
AjaxTree.ROOT_INPUT_CLASS = "treeNodeInput"; // The class for the hidden inputs that contain the root data
AjaxTree.TREE_NODE_CLASS = "treeNode";
AjaxTree.NODE_STRING_NAME = "nodeString";
AjaxTree.CLICK_AREA_CLASS = "clickArea";
AjaxTree.ROOT_OPEN_CLASS = "rootOpen";
AjaxTree.ROOT_CLOSED_CLASS = "rootClosed";
AjaxTree.ROOT_LEAF_CLASS = "rootLeaf";
AjaxTree.NODE_LABEL_CLASS = "nodeLabel";
AjaxTree.USER_DATA_CLASS = "userData";
AjaxTree.TREE_DATA_CLASS = "treeData";
AjaxTree.FILL_DIV_CLASS = "fill";
AjaxTree.PAD_DIV_CLASS = "padDiv";
AjaxTree.PAD_DIV_LAST_CHILD_CLASS = "padDivLastChild";
AjaxTree.PAD_DIV_MID_CHILD_CLASS = "padDivMidChild";
AjaxTree.PAD_DIV_LINE_CLASS = "padDivLine";
AjaxTree.CHILD_WRAPPER_CLASS = "childWrapper";
AjaxTree.SERVER_URL_CLASS = "serverUrl";
AjaxTree.LEAF_STATUS_NAME = "leaf";

AjaxTree.NODE_OPEN_CLASS = "open";
AjaxTree.NODE_CLOSED_CLASS = "closed";
AjaxTree.NODE_LEAF_CLASS = "leaf";

AjaxTree.AJAX_REQUEST_KEY = "tutorial.AJAX_REQUEST";
AjaxTree.CLIENT_ID_REQUEST_KEY = "tutorial.AJAX_CLIENT_ID";
AjaxTree.NODE_STRING_REQUEST_KEY = "tutorial.NODE_STRING";
AjaxTree.REQUEST_TYPE_REQUEST_KEY = "tutorial.REQUEST_TYPE"; // May have many type of requests: expand, delete, create, etc.

AjaxTree.REQUEST_TYPE_EXPAND = "expand";

AjaxTree.XML_RESPONSE_NODE_STRING_TAG = "ns";
/**
 * Load the existing AjaxTree components.  Note this js file must be written only once to the page and is
 * thereafter safe for multiple instances of the tree.
 */
dojo.addOnLoad(AjaxTree.loadTrees);

/*
 nodeData:
    'isLeaf'       : kidIsLeaf,
	'isLastChild'  : lastChild,
	'nodeString'   : kidNodeString,
	'userDataHTML' : userDataHTML,
	'parentNode'   : parentNode,
	'uniqueId'     : kidId,
	'nestingLevel' : nestLevel,
	'ajaxTree'     : AjaxTree,
	'treeNodeDiv'  : domNode,
	'isExpanded'   : boolean,
	'treeDataDiv'  : dom node,
	'userDataDiv'  : dom node,
	'fillDivs'     : node array,
	'padDivs'      : node array,
	'labelDiv'     : dom node
*/